#open "prelude";;
#open "output";;

(*
    This file is the CAML syntactic specification of language ADL
              (Algorithm's Description Language)
*)

type card = All
	| None
	| Le of int
	| Eq of int
	| Odd
	| Even
	| Ge of card * int;;				(* int is the inf *)

let oddp n = if n mod 2 = 1 then true else false;;

let rec simpl_card c = match c with
  Ge(All,l)	-> if l<=0 then All else c
| Ge(None,l)	-> None
| Ge(Le k,l)	-> if k<l then None else if k=l then Eq l else c
| Ge(Eq k,l)	-> if k<l then None else Eq k
| Ge(Odd,l)	-> if l<=1 then Odd else if oddp l then c else Ge(Odd,l+1)
| Ge(Even,l)	-> if l<=0 then Even else if oddp l then Ge(Even,l+1) else c
| Ge(Ge(x,l),m) -> simpl_card (Ge(x,max l m))
| _		-> c;;

let eq_card c1 c2 = simpl_card c1 = simpl_card c2;;

let lowest c = match simpl_card c with
  All		-> 0
| None		-> -1
| Le k		-> 0
| Eq k		-> k
| Odd		-> 1
| Even		-> 0
| Ge(c1,l)	-> l;;

let sub1 c = 			(* decreases by one the nb of elements *)
simpl_card(foo(simpl_card c)) where rec foo = function
  All		-> All
| None		-> None
| Le k		-> if k=0 then None else Le (k-1)
| Eq k		-> Eq (k-1)
| Odd		-> Even
| Even		-> Odd
| Ge(c,l)	-> Ge(foo c,l-1);;

let sub_card a b = if a=b then None else match (a,b) with
  (All,All)		-> None
| (All,Eq 0)		-> Ge(All,1)
| (All,Ge(All,k))	-> if k=1 then (Eq 0) else (Le (k-1))
| x			-> pm "sub_card" x;;

type partial_order = No | Min | Max ;;

type struct = Union of struct * struct
	    | Product of struct * struct
	    | OProduct of struct * struct
	    | Constr of string * struct * card
	    | Sequence of struct * card
	    | Cycle of struct * card
	    | UCycle of struct * card
	    | Set of struct * card
	    | MultiSet of struct * card
	    | Atom of int * int	* string	(* unlabelled *)
	    | LAtom of int * int * string	(* labelled *)
	    | Mark of string
	    | Empty				(* no object *)
	    | Epsilon				(* object of size 0 *)
	    | Emptyset				(* set with 0 element *)
	    | Ident of string * partial_order
;;

(* transforms Sequence(a,c) into Constr("sequence",a,c) *)
let rec tr_struct = function 
  Union(a,b)	-> Union(tr_struct a, tr_struct b)
| Product(a,b)	-> Product(tr_struct a, tr_struct b)
| OProduct(a,b)	-> OProduct(tr_struct a, tr_struct b)
| Constr(p,a,c)	-> Constr(p,tr_struct a,c)
| Sequence(a,c) -> Constr("sequence",tr_struct a,c)
| Cycle(a,c)	-> Constr("cycle",tr_struct a,c)
| UCycle(a,c)	-> Constr("ucycle",tr_struct a,c)
| Set(a,c)	-> Constr("set",tr_struct a,c)
| MultiSet(a,c)	-> Constr("multiset",tr_struct a,c)
| x		-> x
;;

let rec string_of_card = function
  All		-> ""
| None		-> ",card<0"
| Le(k)		-> ",card<=" ^ (son k)
| Eq(k)		-> ",card=" ^ (son k)
| Odd		-> ",card=odd"
| Even		-> ",card=even"
| Ge(c,k)	-> ",card>=" ^ (son k) ^ (string_of_card c)
;;

let string_of_po s = function
  No		-> s
| Min		-> "min(" ^ s ^ ")"
| Max		-> "max(" ^ s ^ ")"
;;

let string_of_struct = foo where rec foo = function
  Union(a,b)	-> (foo a) ^ " | " ^ (foo b)
| Product(a,b)	-> (foo a) ^ " " ^ (foo b)
| OProduct(a,b)	-> (foo a) ^ " " ^ (foo b)
| Constr(p,a,c) -> p ^ "(" ^ (foo a) ^ (string_of_card c) ^ ")"
| Sequence(a,c)	-> "sequence(" ^ (foo a) ^ (string_of_card c) ^ ")"
| Cycle(a,c)	-> "cycle(" ^ (foo a) ^ (string_of_card c) ^ ")"
| UCycle(a,c)	-> "ucycle(" ^ (foo a) ^ (string_of_card c) ^ ")"
| Set(a,c)	-> "set(" ^ (foo a) ^ (string_of_card c) ^ ")"
| MultiSet(a,c)	-> "multiset(" ^ (foo a) ^ (string_of_card c) ^ ")"
| Atom(n,k,s)	-> "atom(" ^ (son n) ^ "," ^ (son k) ^ "," ^ s ^ ")"
| LAtom(n,k,s)	-> "LAtom(" ^ (son n) ^ "," ^ (son k) ^ "," ^ s ^ ")"
| Epsilon	-> "epsilon"
| Ident(s,po)	-> string_of_po s po
| Empty		-> "Empty"
| x		-> pm "string_of_struct" x
;;

let get_def x env = try assoc x env 
		  with _ -> err "get_def" ("variable "^x^" is undefined") ();;

let is_constr s = mem s
 ["product";"sequence";"set";"multiset";"cycle";"ucycle";"diag"];;

let rec transform = function
  ("product",l) -> (match list_length l with
			  0 -> Epsilon
			| 1 -> hd l
			| _ -> if exists 
				(fun (Ident(_,(Max|Min)))->true | _ -> false) l
				then OProduct(hd l,transform("product",tl l))
				else Product(hd l,transform("product",tl l)))
| (x,l)		-> if is_constr x then
			(if list_length l=1 then transform_card(x,hd l,All)
			else user_err "bad number of arguments for constructor"
				x)
		   else transform ("product",(Ident(x,No))::l)

and transform_card = function
  ("sequence",a,c)  -> Sequence(a,c)
| ("cycle",a,c)	    -> Cycle(a,c)
| ("set",a,c)	    -> Set(a,c)
| ("multiset",a,c)  -> MultiSet(a,c)
| ("ucycle",a,c)    -> UCycle(a,c)
| (s,a,c)	    -> if is_constr s then Constr(s,a,c)
		       else if c=All then
			transform("product",[(Ident(s,No));a])
		       else user_err (s^" is not a list constructor, no cardinality restriction allowed") (string_of_card c)
;;

let rec untransform = function
  Product(a,b) -> a::untransform b
| OProduct(a,b) -> a::untransform b
| x	       -> [x];;

let tr_prod l = transform("product",map (fun id -> Ident(id,No)) l);;

type type_decl == string * struct;;

let types = ref [("",Empty)];;

type cost =       NumCost of int
                | ParCost of string;; 

type meas_decl == string * cost ;; 

let measures = ref [("",NumCost 0)];;

type pattern = Pat_id of string
	     | Pat_all
	     | Pat_epsilon
	     | Pat_emptyset
	     | Pat_prod of pattern * pattern ;;

let rec transform1 l = match list_length l with
			  0 -> Pat_epsilon
			| 1 -> hd l
			| _ -> Pat_prod(hd l,transform1(tl l));;

type expression = Exp_id of string
		| Exp_constr of string * expression list
		| Call of string * expression list ;;

let cc = end_it_list (fun x y -> x^","^y) "";;

let boundp s env = try get_def s env; true
                   with Failure _ -> false;;

let rec pp_exp = function
  Exp_id s	  		-> s
| Exp_constr("product",(Exp_id s)::l)-> if boundp s !types & list_length l>=1
				   then s ^ "(" ^ cc(map pp_exp l) ^ ")"
				   else "product(" ^ 
					cc(map pp_exp ((Exp_id s)::l)) ^ ")"
| Exp_constr(c,l)		-> c ^ "(" ^ cc(map pp_exp l) ^ ")"
| Call(f,l)	  		-> f ^ "(" ^ cc (map pp_exp l) ^ ")";;

type condition  = Equal of struct * pattern 
	        | Predicate of string * string
;;

type repeat	= RepNum of int
		| RepId of string
;;

type AdlInstr = Cases of string list * (pattern list * AdlInstr) list
		| Casetype of string * (pattern list * AdlInstr) list
		| Instructions of AdlInstr list
		| IfThenElse of condition * AdlInstr * AdlInstr
		| Forall of string * string * AdlInstr
		| Forone of string * string * AdlInstr
		| Expression of expression
		| Repeat of AdlInstr * repeat
		| Empty_instr ;;

type function_decl == string * (type_decl list * (string * AdlInstr)) ;;

let functions = ref [("",([("",Empty)],("",Instructions [])))];;

type adl_file = File of (type_decl list) * (function_decl list) * (meas_decl list);;

let rec distrib y = function
      []    -> []
    | x::tl -> (x,y) :: (distrib y tl) ;;

let add_string s1 s2 = match (s1,s2) with
  ("0","0") -> "0"
| ("0",s)   -> s
| (s,"0")   -> s
| _         -> s1^"+"^s2;;

let rec level c = function
  []     -> 1
| "("::y -> (level c y)+1
| ")"::y -> (level c y)-1
| x::y   -> let l = level c y in if mem x c then min 0 l else l;;

let mul_string s1 s2 = if s1="0" or s2="0" then "0" 
		       else "(" ^ s1 ^ ")*(" ^ s2 ^ ")";;

let marks = ref ["sjkdsdf"];;

let add_mark s = if not (mem s !marks) then marks:=s::!marks;();;

