#open "prelude";;
#open "output";;
#open "adl";;
#open "types";;
#open "cite";;

(* these functions checks that the data type specification is well defined *)

type valuation = NumVal of int | Infinity;;

let string_of_val v = match v with
  Infinity	-> "infinity"
| NumVal(n)	-> string_of_int n
;;

let MinVal a b = match (a,b) with
  (Infinity,x)	-> x
| (y,Infinity)	-> y
| (NumVal(n),NumVal(m))	-> NumVal(min n m)
;;

let MaxVal a b = match (a,b) with	(* maximum <> Infinity *)
  (Infinity,x)	-> x
| (y,Infinity)	-> y
| (NumVal(n),NumVal(m))	-> NumVal(max n m)
;;

let AddVal a b = match (a,b) with
  (Infinity,x)	-> Infinity
| (y,Infinity)	-> Infinity
| (NumVal(n),NumVal(m))	-> NumVal(add_int n m)
;;

let MulVal a k = if k=0 then NumVal(0) else match a with
  Infinity	-> Infinity
| NumVal(n)	-> NumVal(mult_int n k)
;;

(* cf [Zimmermann91,p.26] *)
let properties = (* list of (Phi, f_Phi, "2a" | "2b", (P|Q)* ) *)
	[("sequence", ((fun n c -> MulVal n c),	("2a", ["P"])));
	 ("multiset", ((fun n c -> MulVal n c),	("2a", ["P"])));
	 ("set",      ((fun n c -> MulVal n c),	("2b", ["P"])));
	 ("cycle",    ((fun n c -> MulVal n (max 1 c)),	("2a", ["P"])));
	 ("ucycle",   ((fun n c -> MulVal n (max 1 c)),	("2a", ["P"])));
	 ("diag",     ((fun n c -> MulVal n 2),	("2b", ["P";"Q"])))];;

let property1 p = fst(assoc p properties);;
let property2 p = fst(snd(assoc p properties));;
let property3 p = snd(snd(assoc p properties));;	(* list of "P" | "Q" *)

let valuation symbvals = foo where rec foo st = match st with
  Union(a,b)	-> MinVal (foo a) (foo b)
| Product(a,b)	-> AddVal (foo a) (foo b)
| OProduct(a,b)	-> AddVal (foo a) (foo b)
| Sequence(a,c)	-> MulVal (foo a) (lowest c)
| Cycle(a,c)	-> MulVal (foo a) (max 1 (lowest c))
| UCycle(a,c)	-> MulVal (foo a) (max 1 (lowest c))
| Set(a,c)	-> let lc = lowest c in
		   let res = MulVal (foo a) lc in
		     if lc>=2 then war ("the valuation of "^string_of_struct 
		     st^" may be greater than "^string_of_val res); res
| MultiSet(a,c)	-> MulVal (foo a) (lowest c)
| Atom(n,k,s)	-> NumVal(n)
| LAtom(n,k,s)	-> NumVal(n)
| Mark(s)	-> NumVal(0)
| Empty		-> Infinity
| Epsilon	-> NumVal(0)
| Emptyset	-> NumVal(0)
| Ident(s,po)	-> (try assoc s symbvals 
		   with _ -> user_err "undefined symbol" s)
| Constr(p,a,c)	-> property1 p (foo a) (lowest c)
;;

let vals = ref [("toto",NumVal(0))];;

(* this function implements algorithm A of the article 
	Automatic Average-case Analysis of Algorithms *)
let rec compute_valuations symbols symbvals =
 let newsymbvals = map
  (fun s -> let rm = assoc s !types in (s,valuation symbvals rm))
  symbols in
 if newsymbvals <> symbvals then compute_valuations symbols newsymbvals
 else vals:=symbvals
;;

let print_valuations () = map (fun (s,v) -> if not atomp s then 
	rap 5 ("     The valuation of "^s^" is "^(string_of_val v))) !vals;
rap 5 ""
;;

let val x = valuation !vals x;;

let is_non_terminal s = let rm = assoc s !types in match rm with
  Atom(_,_,_)
| LAtom(_,_,_)
| Mark(_)
| Empty
| Epsilon
| Emptyset	-> false
| _		-> true
;;

let number_of_non_terminals () =
let l = map (fun (s,_) -> if is_non_terminal s then 1 else 0) !types
in end_it_list add_int 0 l
;;

(* Union(x,Union(y,z)) -> [x;y;z] *)
let rec flatunion st = match st with
  Union(a,b)	-> (flatunion a) @ (flatunion b)
| _		-> [st]
;;

let productions s = let rm = assoc s !types in flatunion rm;;

(* this function returns the list of words derived from e in one step *)
let rec set_of_derived e = match e with
  Union(a,b)	-> union (set_of_derived a) (set_of_derived b)
| Product(a,b)	-> let sa = map (fun x -> Product(x,b)) (set_of_derived a)
                   and sb  = map (fun x -> Product(a,x)) (set_of_derived b)
                   in union sa sb
| OProduct(a,b)	-> let sa = map (fun x -> OProduct(x,b)) (set_of_derived a)
                   and sb  = map (fun x -> OProduct(a,x)) (set_of_derived b)
                   in union sa sb
| Sequence(a,c)	-> map (fun x -> Sequence(x,c)) (set_of_derived a)
| Cycle(a,c)	-> map (fun x -> Cycle(x,c)) (set_of_derived a)
| UCycle(a,c)	-> map (fun x -> UCycle(x,c)) (set_of_derived a)
| Set(a,c)	-> map (fun x -> Set(x,c)) (set_of_derived a)
| MultiSet(a,c)	-> map (fun x -> MultiSet(x,c)) (set_of_derived a)
| Atom(_,_,_)
| LAtom(_,_,_)
| Mark(_)
| Empty
| Epsilon
| Emptyset	-> [e]
| Ident(s,po)	-> productions s
| xx		-> pm "set_of_derived" xx
;;

let check_step2 = foo where rec foo st = match st with
  Union(a,b)	-> foo a; foo b
| Product(a,b)	-> foo a; foo b
| OProduct(a,b) -> foo a; foo b
| Constr(p,a,c)	-> if property2 p = "2a" & val a=NumVal(0) & lowest c = 0 then
		     nwf ("this production derives plenty objects of size 0: "
			^string_of_struct st);
		   foo a
| _		-> true
;;

let rec search_constructors = function
  Union(a,b)	-> union (search_constructors a) (search_constructors b)
| Product(a,b)	-> union (search_constructors a) (search_constructors b)
| OProduct(a,b)	-> union (search_constructors a) (search_constructors b)
| Constr(p,a,_)	-> union [p] (search_constructors a)
| _		-> []
;;

(* compute_graph n returns the graph of the relation \prec_n defined in 
   [Zimmermann91,p. 31] in a form of a list of pairs : A \prec_n B ==> (A,B) *)
let rec cga_aux n = foo where rec foo = function
  Union(a,b)	-> union (foo a) (foo b)
| Product(a,b)	-> union (if val a = NumVal(0) then foo b else [])
			 (if val b = NumVal(0) then foo a else [])
| OProduct(a,b)	-> union (if val a = NumVal(0) then foo b else [])
			 (if val b = NumVal(0) then foo a else [])
| Ident(s,_)	-> [s]
| Constr(p,a,_)	-> let p3=property3 p in 
			if nth p3 (min ((list_length p3)-1) n)="P"
			then foo a else []
| Epsilon
| Empty
| Atom(_)
| LAtom(_)	-> []
| xx		-> pm "cga_aux" xx
;;

let compute_graph n = flat(map compute_graph_aux !types)
where compute_graph_aux (s,t) = map (fun x -> (x,s)) (cga_aux n (tr_struct t))
;;

let unions = it_list union [];;

(* g is the initial graph ( list of pairs (x,y) where x -> y )
   gstar the current transitive closure ( list of triples (x,z,y) where
	x -*-> z and z -> y ) *)
let rec hc_aux g gstar =
 let l = filter_pos (fun [x;z;y] -> x=y | _ -> snh 2) gstar in
  if l<>[] then hd l 	(* there is a cycle *)
  else			(* there is no cycle for now *)
   let new_gstar = unions (map (fun [x;_;y] -> flat
    (map (fun (z,t) -> if z=y then [[x;y;t]] else []) g)|_->snh 2) gstar) in
   if new_gstar=[]	(* no new relation *)
   then []		(* means no cycle *)
   else match hc_aux g new_gstar with
     []		-> []
   | x::z::r	-> (hd (filter_pos (fun [a;_;b] -> a=x & b=z | _ -> snh 2) gstar))@r
;;

(* l is a list of pairs *)
let rec find_cycle l = match hc_aux l (map (fun (x,y) -> [x;x;y]) l) with
  [] -> []
| l  -> tl l;;

let look_for_loops n = 	(* check for loops for size n *)
      let c = find_cycle (compute_graph n) in
      if c<>[] then nwf ("there is a cycle for size "^son n^": "
	^end_it_list (fun x y -> x^" < "^y) "" c);;

(* this function implements algorithm B of [Zimmermann91] *)
let rec check_well_defined () =
(* step 1 : check if all valuations are finite *)
   let l = filter_pos (fun(s,v)->v=Infinity) !vals in
(*      types := map (fun (s,t) -> (s,tr_struct t)) !types; *)
      if l<>[] then war ("infinite valuation for "^(cc(map fst l)));
(* step 2 : check for productions X -> Phi(Y) where val(Y)=0 and
	Phi such that \forall Y, val(Y)=0 ==> card(Phi(Y))=\infty *)
   let l = map (fun y -> tr_struct (snd y)) !types in
      map check_step2 l;
(* computing N = max(N_Phi) *)
      let cnstrs = make_set(flat(map search_constructors l))
      and maxval = (match it_list MaxVal (NumVal 0) (map snd !vals) with
		NumVal i -> i | _ -> err "check_well_defined" "Infinity" ())
      and minval = (match it_list MinVal Infinity (map snd !vals) with
		NumVal i -> i | _ -> err "check_well_defined" "Infinity" ())
	in 
	rap 5 ("     Minimal valuation is "^son minval^
		",     maximal valuation is "^son maxval^".");
	let N = it_list max maxval 
		(map (fun x -> list_length x -1) (map property3 cnstrs))
	   in rap 5 ("     N="^son N);
(* steps 3-4 : look for loops for minval <= n < N *)
      do_list look_for_loops (interval minval (N-1));
(* step 5 : look for loops for n=N *)
      look_for_loops N;
      mess 3 "   The data type specification is well-defined."
;;

(* patch for bug in Caml-Light 0.74 compiler *)
let unions = it_list union [];;

let Cga_aux args = foo where rec foo = function
  Cases(_,l)		-> unions (map foo (map snd l))
| Casetype(_,l)		-> unions (map foo (map snd l))
| Instructions(l)	-> unions (map foo l)
| IfThenElse(_,t,e)	-> union (foo t) (foo e)
| Forall(_,_,i)		-> foo i
| Forone(_,_,i)		-> foo i
| Repeat(i,_)		-> foo i
| Empty_instr		-> []
| Expression(e)		-> (match e with
			     Call(g,l) -> if args=map pp_exp l then [g] else []
			   | Exp_id(_) -> []
			   |Exp_constr(_,l)->unions(map 
				foo(map (fun x -> Expression x) l)))
;;

(* for procedures *)
let Compute_graph () = flat(map Compute_graph_aux !functions)
where Compute_graph_aux (f,(args,(res,b))) = 
	let largs = map fst args in
	let l = intersect largs (map fst !types) in
        if l <> [] then user_err "data types and function parameters should have\n\tdifferent names" (cc l);
	map (fun x -> (x,f)) (Cga_aux largs b)
;;

let check_procedures () = 
      let c = find_cycle (Compute_graph ()) in
      if c<>[] then user_err 
("The program is not well defined: there is a loop between procedures: "^
	end_it_list (fun x y -> x^" < "^y) "" c) ""
;;

let check_all () = 
mess 3 "   Checking type declarations ...";
mess 4 "    Computing valuations [Zimmermann91, algorithm A p. 28] ...";
if !printlevel>=4 then cite "Zimmermann91";
let symbols = map fst !types
and nbnt = number_of_non_terminals ()
in compute_valuations symbols (map (fun s -> (s,Infinity)) symbols);
 print_valuations ();
 mess 4 "    Checking for well-definedness of the data type specification [Zimmermann91, algorithm B p. 33]:";
 check_well_defined ();
 mess 3 "   Checking procedures ...";
 check_procedures ()
;;
