#open "prelude";;
#open "output";;
#open "adl";;
#open "gf";;

(*********************** notations *********************************

    Q(z) = 1/(1-z)					pseudo-inverse
    L(z) = log(1/(1-z))
    P(a,z) = a(z) - a(z^2)/2 + a(z^3)/3 - ...		Polya operator
    texp(x,k) = 1 + x + x^2/2 + ... + x^k/k!		truncated exponential
*)
(*********************** generating functions **********************)

let gf_sequence gftype A c = match c with        (* doesn't depend on gftype *)

  All	    -> Op("Q",A)
| Eq(k)	    -> Power(A,k)
| Le(k)	    -> Ratio(OneMinus(Power(A,k+1)),OneMinus(A))
| Odd	    -> Times(A,Op("Q",Power(A,2)))
| Even	    -> Op("Q",Power(A,2))
| Ge(All,k) -> Times(Power(A,k),Op("Q",A))
| None	    -> Gf_num(0)
| _	    -> pm "gf_sequence" "other type of card" ;;

let rec gf_cycle gftype A c = if gftype = "ordinary" 
then nyi "gf_cycle" "ordinary" else match c with

  All	    -> Op("L",A)
| Eq(k)	    -> NRatio(Power(A,k),k)
| Le(k)	    -> Pluss(map (fun j -> NRatio(Power(A,j),j)) (range k))
| Odd	    -> NRatio(Op("log",Ratio(OnePlus(A),OneMinus(A))),2)
| Even	    -> NRatio(Op("L",Power(A,2)),2)
| None	    -> Gf_num(0)
| Ge(All,k) -> if k<=1 then Op("L",A)
		else  Minus(Op("L",A),gf_cycle gftype A (Le(k-1)))
| _	    -> pm "gf_cycle" "other type of card" ;;

let rec gf_ucycle gftype A c = if gftype = "ordinary" 
then nyi "gf_ucycle" "ordinary" else match c with

  All	    -> NRatio(Plus(Op("L",A),Plus(A,NRatio(Power(A,2),2))),2)
| Eq(k)	    -> if k=1 then A else if k=2 then NRatio(Power(A,2),2)
	       else NRatio(Power(A,k),2*k)
| Le(k)	    -> if k=1 then A else if k=2 then Plus(A,NRatio(Power(A,2),2))
	       else Pluss([A;NRatio(Power(A,2),2)] @ 
		(map (fun j -> NRatio(Power(A,j),2*j)) (interval 3 k)))
| Ge(All,k) -> if k<=1 then gf_ucycle gftype A All
	       else if k=2 
		then NRatio(Plus(Minus(Op("L",A),A),NRatio(Power(A,2),2)),2)
	       else if k=3 
		then NRatio(Minus(Op("L",A),Plus(A,NRatio(Power(A,2),2))),2)
	       else nyi "gf_ucycle" "ge(k) for k>=4"
| Odd	    -> Plus(NRatio(Op("log",Ratio(OnePlus(A),OneMinus(A))),4),
		    NRatio(A,2))
| Even	    -> NRatio(Op("log",Plus(Op("Q",Power(A,2)),Power(A,2))),4)
| None	    -> Gf_num(0)
| _	    -> pm "gf_ucycle" "other type of card" ;;

let gf_set gftype A c = 
if gftype = "ordinary" then (match c with

      All	    -> Polya(A,-1)
    | Eq(k)	    -> Polya(A,k)
    | Le(k)	    -> nyi "gf_set" "card <= k for ogf"
    | Odd	    -> nyi "gf_set" "card = odd for ogf"
    | Even	    -> nyi "gf_set" "card = even for ogf"
    | None	    -> Gf_num(0)
    | _		    -> pm "gf_set" "other type of card")

else if gftype = "exponential" then (match c with

      All	    -> Op("exp",A)
    | Eq(k)	    -> Ratio(Power(A,k),Fact(k))
    | Le(k)	    -> Texp(A,k)
    | Odd	    -> NRatio(Plus(Op("exp",A),Op("exp",UMinus(A))),2)
    | Even	    -> Op("cosh",A)
    | Ge(All,k)     -> if k=1 then Minus(Op("exp",A),One)
		       else if k=2 then Minus(Op("exp",A),OnePlus(A))
		       else Minus(Op("exp",A),Texp(A,k-1))
    | None	    -> Gf_num(0)
    | _		    -> pm "gf_set" "other type of card")

else err "gf_set" "type is neither ordinary nor exponential" gftype;;

let gf_multiset gftype A c = 
if gftype = "ordinary" then (match c with

      All	    -> MPolya(A,-1)
    | Eq(k)	    -> MPolya(A,k)
    | Ge(All,k)	    -> if k=1 then Minus(MPolya(A,-1),One)
			else nyi "gf_multiset" "Ge(All,k) for k>=2"
    | Le(k)	    -> nyi "gf_multiset" "card <= k for ogf"
    | Odd	    -> nyi "gf_multiset" "card = odd for ogf"
    | Even	    -> nyi "gf_multiset" "card = even for ogf"
    | None	    -> Gf_num(0)
    | _		    -> err "gf_multiset" "other type of card" c)

else if gftype = "exponential" then 
	err "gf_multiset" "multisets exists only for unlabelled objects" A
else err "gf_multiset" "type is neither ordinary nor exponential" gftype;;

let gf_constr p gftype A c = match p with
  "diag"	-> Op("Diag",A)
| _ -> pm "gf_constr" p;;

let gf_atom gftype n multiplicity s = 
let x =
(if gftype = "ordinary" then
		   if n=0 then One
		   else if n=1 then Gf_par(s)
                   else Power(Gf_par(s),n)
else if gftype = "exponential" then
		   if n=0 then One
		   else if n=1 then Gf_par(s)
                   else Ratio(Power(Gf_par(s),n),Fact(n))
else err "gf_atom" "type is neither ordinary nor exponential" gftype)
in if multiplicity=1 then x else Times(Gf_num multiplicity,x);;

(*****************************************************************************)

let compl_forall_set gftype tq A c = 
if gftype = "ordinary" then nyi "compl_forall_set" "unlabelled structure"
else if gftype = "exponential" then (match c with

      All	-> Times(Op("exp",A),tq)
    | Eq k	-> Times(tq,Ratio(Power(A,k-1),Fact(k-1)))
    | Ge(All,k)	-> if k=1 then Times(Op("exp",A),tq)
		   else nyi "compl_forall_set" "exponential: Ge(All,k) for k>1"
    | _		-> nyi "compl_forall_set" "exponential : card other than All or Eq")
else err "compl_forall_set" "type is neither ordinary nor exponential" gftype;;

let compl_forall_multiset gftype tq A c = 
if gftype = "exponential" then err "compl_forall_multiset"
	"in labelled universe, multisets reduce to sets" ()
else if gftype = "ordinary" then (match c with

      All	-> Times(MPolya(A,-1),Op("prog_MP",tq))
    | _		-> nyi "compl_forall_multiset" "card other than All")
else err "compl_forall_multiset" "type is neither ordinary nor exponential" gftype;;

let compl_forall_cycle gftype tq A c = 
if gftype = "ordinary" then nyi "compl_forall_cycle" "unlabelled structure"
else if gftype = "exponential" then (match c with

      All	-> Ratio(tq,OneMinus(A))
    | _		-> nyi "compl_forall_cycle" "exponential : card other than All")
else err "compl_forall_cycle" "type is neither ordinary nor exponential" gftype;;

let compl_forall_ucycle gftype tq A c = 
if gftype = "ordinary" then nyi "compl_forall_ucycle" "unlabelled structure"
else if gftype = "exponential" then (match c with

      All	-> Times(NRatio(tq,2),Plus(Plus(One,A),Ratio(One,OneMinus(A))))
    | _		-> nyi "compl_forall_ucycle" "exponential : card other than All")
else err "compl_forall_ucycle" "type is neither ordinary nor exponential" gftype;;

let compl_forall_sequence gftype tq A c =
if gftype = "ordinary" or gftype = "exponential" then (match c with

      All       -> Times(tq,Power(Op("Q",A),2))
    | Eq k	-> Times(Gf_num(k),Times(tq,Power(A,k-1)))
    | Ge(All,1)	-> Times(tq,Power(Op("Q",A),2))
    | Ge(All,2) -> Times(tq,Minus(Power(Op("Q",A),2),One))
    | Le k	-> Times(tq,Tder(A,k))
    | _         -> nyi "compl_forall_sequence" "card other than All")
else err "compl_forall_sequence" 
         "type is neither ordinary nor exponential" gftype;;

let compl_forone_set gftype tq A c = 
if gftype = "ordinary" then nyi "compl_forone_set" "unlabelled structure"
else if gftype = "exponential" then (match c with
      All       -> Times(Ratio(Minus(Op("exp",A),One),A),tq)
    | _         -> nyi "compl_forone_set" "exponential : card other than All")
else err "compl_forone_set" "type is neither ordinary nor exponential" gftype;;

let compl_forone_cycle gftype tq A c = 
if gftype = "ordinary" then nyi "compl_forone_cycle" "unlabelled structure"
else if gftype = "exponential" then (match c with
      All	-> Times(tq,Ratio(Op("L",A),A))
    | _		-> nyi "compl_forone_cycle" "exponential : card other than All")
else err "compl_forone_cycle" "type is neither ordinary nor exponential" gftype;;

let compl_forone_sequence gftype tq A c =
if gftype = "ordinary" or gftype = "exponential" then (match c with
      All       -> Times(tq,Op("Q",A))
    | _         -> nyi "compl_forone_sequence" "card other than All")
else err "compl_forone_sequence" 
         "type is neither ordinary nor exponential" gftype;;

