#open "sys";;
#open "io";;
#open "prelude";;
#open "output";;
#open "maple";;
#open "report";;

exception Quit;;
exception BadNumberOfArgs of string;;
exception BadTypeOfArgs of string;;
exception UnknownCommand of string;;

let input_stream = ref std_in;;

let print_prompt() =
  print_string "luo> "; flush std_out
;;

let read_fun =
  let bol = ref true in
fun() ->
  if !bol then print_prompt();
  let c = input_char !input_stream in
  if !input_stream != std_in then print_char c;
  bol := c == `\n`;
  c
;;

type arg = IntArg of int | IdArg of string;;

type command = Command of string * arg list
	     | Help of string;;

type token_type = IDENT of string | INT of int | ENDCOMMAND | 
	STRING of string | HELP of string;;

let buff = create_string 1000;;

let rec ident len = function
  [< '(`a`..`z` | `A` .. `Z` | `0` .. `9` | `_` | `'`) as c; s >] ->
    set_nth_char buff len c; ident (succ len) s
| [< >] -> sub_string buff 0 len
;;

let rec get_string len = function
  [< '`"` >] -> sub_string buff 0 len
| [< 'c; s >] -> set_nth_char buff len c; get_string (succ len) s
;;

let rec number n = function
  [< '`0` .. `9` as d; s >] ->
    number(10*n+int_of_char d-int_of_char`0`) s
| [< >] -> n
;;

let args = function
  [< 'IDENT s >] -> IdArg s
| [< 'STRING s >] -> IdArg s
| [< 'INT n >] -> IntArg n
;;

let top = function
  [< 'IDENT s; (list args) l; 'ENDCOMMAND >] -> Command(s,l)
| [< 'HELP s >] -> Help s
;;

let rec next_token = function
  [< '(`a`..`z` | `A` .. `Z`) as c; s >] ->
    set_nth_char buff 0 c; IDENT(ident 1 s)
| [< '`0` .. `9` as d; s >] ->
    INT(number (int_of_char d-int_of_char `0`) s)
| [< '` ` | `\n` | `\t`; s >] -> next_token s
| [< '`;` >] -> ENDCOMMAND
| [< '`?`; s >] -> token_help s
| [< '`"`; s >] -> STRING(get_string 0 s)
| [< 'x >] -> raise Parse_error

and token_help = function
  [< '(`a`..`z` | `A` .. `Z`) as c; s >] ->
    set_nth_char buff 0 c; HELP(ident 1 s)
| [< >] -> HELP("")
;;

let rec reset_lexer = function
  [< '`\n` >] -> ()
| [< '_; reset_lexer _ >] -> ()
| [< >] -> ()
;;

let luo_interp l s = match s with
  "asympt" -> if list_length l <> 1 then raise (BadNumberOfArgs s)
	      else (match hd l with
		  IntArg(_) -> raise (BadTypeOfArgs s)
		| IdArg(f) -> if !gtype="" then
				failwith "no current Adl program"
			      else asympt f)
| "distribution" -> if list_length l <> 2 then raise (BadNumberOfArgs s)
              else (match l with
                [IdArg(m);IdArg(s)] -> distribution m s
              | _ -> raise (BadTypeOfArgs s))
| "evalf" -> if list_length l <> 1 then raise (BadNumberOfArgs s)
	    else (match hd l with
		  IntArg(_) -> raise (BadTypeOfArgs s)
		| IdArg("true") -> evalf:=true
		| IdArg("false") -> evalf:=false
		| IdArg(f) -> raise (BadTypeOfArgs f))
| "load" -> if list_length l <> 1 then raise (BadNumberOfArgs s)
	    else (match hd l with
		  IntArg(_) -> raise (BadTypeOfArgs s)
		| IdArg(f) -> load_adl f; alas_and_solver())
| "moment" -> if list_length l <> 3 then raise (BadNumberOfArgs s)
	      else (match l with
		[IdArg(m);IdArg(s);IntArg(k)] -> moment m s k
	      | _ -> raise (BadTypeOfArgs s))
| "order" -> if list_length l <> 1 then raise (BadNumberOfArgs s)
		else (match hd l with
		  IdArg(_) -> raise (BadTypeOfArgs s)
		| IntArg(n) -> order:=n)
| "prettyprint" -> if list_length l <> 1 then raise (BadNumberOfArgs s)
	    else (match hd l with
		  IntArg(_) -> raise (BadTypeOfArgs s)
		| IdArg(f)->maple_eval_quiet ("interface(prettyprint="^f^"):"))
| "mapleeval"  -> if list_length l <> 1 then raise (BadNumberOfArgs s)
		else (match hd l with
		  IntArg(_) -> raise (BadTypeOfArgs s)
		| IdArg(s) -> maple_eval_quiet s)
| "printlevel" -> if list_length l <> 1 then raise (BadNumberOfArgs s)
		  else (match hd l with
		  IdArg(_) -> raise (BadTypeOfArgs s)
		| IntArg(n) -> printlevel:=n;
		      maple_eval_quiet ("set_luo_level("^string_of_int n^"):"))
| "quit" -> raise Quit
| "variance" -> if list_length l <> 2 then raise (BadNumberOfArgs s)
              else (match l with
                [IdArg(m);IdArg(s)] -> variance m s
              | _ -> raise (BadTypeOfArgs s))
| _ -> raise (UnknownCommand s)
;;

let go() =
  print_string "                ___    Lambda-Upsilon-Omega 2.0\n";
  print_string "   /\\   \\   /  /   \\   Bruno Salvy & Paul Zimmermann, Institut National\n";
  print_string "  /  \\   \\ /  |     |  de Recherche en Informatique et en Automatique\n";
  print_string "_/    \\_  |   _\\   /_  Type ? for help\n";
  print_string "          |            \n";
  try
    let cstrm = stream_from read_fun in
    let strm = stream_from (fun _ -> next_token cstrm) in
    while true do
      try
        match top strm with
          Command(s,l) -> luo_interp l s
        | Help(s) -> help s
      with
        Parse_failure ->
          (* print_newline(); raise Quit *)
          print_newline();
          print_string "*** Parsing error."; print_newline();
          reset_lexer cstrm; stream_next strm; ()
      | Parse_error ->
          print_newline();
          print_string "*** Syntax error."; print_newline();
          reset_lexer cstrm; stream_next strm; ()
      | UnknownCommand s ->
          print_newline();
          print_string "*** Unknown command: "; print_string s; print_newline();
          print_string "    The available commands can be obtained with ?<return>";
          print_newline()
      | Failure s ->
          print_newline();
          print_string "*** "; print_string s; print_newline()
      | BadNumberOfArgs(s) ->
	  print_newline();
          print_string "*** Bad number of arguments for command ";
	  print_string s; print_newline()
      | BadTypeOfArgs(s) ->
	  print_newline();
          print_string "*** Bad type of arguments for command ";
	  print_string s; print_newline()
    done
  with
    Quit -> mapleput "quit();"; print_string "See you later.\n"; flush std_out
  | Failure s ->
      print_string "*** Failed: "; print_string s; print_newline()
;;

